==> emu_mps-1.2.6.tar.gz: emu_mps-1.2.6/emu_mps/tdvp.py <==
import torch

from emu_base import krylov_exp
from emu_mps import MPS, MPO
from emu_mps.utils import split_tensor
from emu_mps.mps_config import MPSConfig


def new_right_bath(
    bath: torch.Tensor, state: torch.Tensor, op: torch.Tensor
) -> torch.Tensor:
    bath = torch.tensordot(state, bath, ([2], [2]))
    bath = torch.tensordot(op.to(bath.device), bath, ([2, 3], [1, 3]))
    bath = torch.tensordot(state.conj(), bath, ([1, 2], [1, 3]))
    return bath


"""
function to compute the right baths. The three indices in the bath are as follows:
(bond of state conj, bond of operator, bond of state)
The baths have shape
-xx
-xx
-xx
with the index ordering (top, middle, bottom)
bath tensors are put on the device of the factor to the left
"""


def right_baths(state: MPS, op: MPO, final_qubit: int) -> list[torch.Tensor]:
    state_factor = state.factors[-1]
    bath = torch.ones(1, 1, 1, device=state_factor.device, dtype=state_factor.dtype)
    baths = [bath]
    for i in range(len(state.factors) - 1, final_qubit - 1, -1):
        bath = new_right_bath(bath, state.factors[i], op.factors[i])
        bath = bath.to(state.factors[i - 1].device)
        baths.append(bath)
    return baths


"""
Computes H(psi) where
    x-    -x
    x  ||  x             ||
H = x- xx -x  and psi = -xx-
    x  ||  x
    x-    -x

Expects the two qubit factors of the MPS precontracted,
with one 'fat' physical index of dim 4 and index ordering
(left bond, physical index, right bond):
         ||
      -xxxxxx-
The Hamiltonian should have an index ordering of
(left bond, out, in, right bond).
The baths must have shape (top, middle, bottom).
All tensors must be on the same device
"""


def apply_effective_Hamiltonian(
    state: torch.Tensor,
    ham: torch.Tensor,
    left_bath: torch.Tensor,
    right_bath: torch.Tensor,
) -> torch.Tensor:
    assert left_bath.ndim == 3 and left_bath.shape[0] == left_bath.shape[2]
    assert right_bath.ndim == 3 and right_bath.shape[0] == right_bath.shape[2]
    assert left_bath.shape[2] == state.shape[0] and right_bath.shape[2] == state.shape[2]
    assert left_bath.shape[1] == ham.shape[0] and right_bath.shape[1] == ham.shape[3]

    # the optimal contraction order depends on the details
    # this order seems to be pretty balanced, but needs to be
    # revisited when use-cases are more well-known
    state = torch.tensordot(left_bath, state, 1)
    state = torch.tensordot(state, ham, ([1, 2], [0, 2]))
    state = torch.tensordot(state, right_bath, ([3, 1], [1, 2]))
    return state


_TIME_CONVERSION_COEFF = 0.001  # Omega and delta are given in rad/ms, dt in ns


def evolve_pair(
    *,
    state_factors: list[torch.Tensor],
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factors: list[torch.Tensor],
    dt: float,
    orth_center_right: bool,
    is_hermitian: bool,
    config: MPSConfig,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Time evolution of a pair of tensors of a tensor train using baths and truncated SVD.
    Returned state tensors are kept on their respective devices.
    """
    assert len(state_factors) == 2
    assert len(baths) == 2
    assert len(ham_factors) == 2

    left_state_factor, right_state_factor = state_factors
    left_bath, right_bath = baths
    left_ham_factor, right_ham_factor = ham_factors

    left_device = left_state_factor.device
    right_device = right_state_factor.device

    left_bond_dim = left_state_factor.shape[0]
    right_bond_dim = right_state_factor.shape[-1]

    # Computation is done on left_device (arbitrary)

    combined_state_factors = torch.tensordot(
        left_state_factor, right_state_factor.to(left_device), dims=1
    ).reshape(left_bond_dim, 4, right_bond_dim)

    left_ham_factor = left_ham_factor.to(left_device)
    right_ham_factor = right_ham_factor.to(left_device)

    combined_hamiltonian_factors = (
        torch.tensordot(left_ham_factor, right_ham_factor, dims=1)
        .transpose(2, 3)
        .reshape(left_ham_factor.shape[0], 4, 4, -1)
    )

    op = (
        lambda x: -_TIME_CONVERSION_COEFF
        * 1j
        * dt
        * apply_effective_Hamiltonian(
            x, combined_hamiltonian_factors, left_bath, right_bath
        )
    )

    evol = krylov_exp(
        op,
        combined_state_factors,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    ).reshape(left_bond_dim * 2, 2 * right_bond_dim)

    l, r = split_tensor(
        evol,
        max_error=config.precision,
        max_rank=config.max_bond_dim,
        orth_center_right=orth_center_right,
    )

    return l.reshape(left_bond_dim, 2, -1), r.reshape(-1, 2, right_bond_dim).to(
        right_device
    )


def evolve_single(
    *,
    state_factor: torch.Tensor,
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factor: torch.Tensor,
    dt: float,
    is_hermitian: bool,
    config: MPSConfig,
) -> torch.Tensor:
    """
    Time evolution of a single tensor of a tensor train using baths.
    """
    assert len(baths) == 2

    left_bath, right_bath = baths

    op = (
        lambda x: -_TIME_CONVERSION_COEFF
        * 1j
        * dt
        * apply_effective_Hamiltonian(
            x,
            ham_factor,
            left_bath,
            right_bath,
        )
    )

    return krylov_exp(
        op,
        state_factor,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    )

==> emu_mps-2.0.4.tar.gz: emu_mps-2.0.4/emu_mps/tdvp.py <==
import torch

from emu_base import krylov_exp
from emu_mps import MPS, MPO
from emu_mps.utils import split_tensor
from emu_mps.mps_config import MPSConfig


def new_right_bath(
    bath: torch.Tensor, state: torch.Tensor, op: torch.Tensor
) -> torch.Tensor:
    bath = torch.tensordot(state, bath, ([2], [2]))
    bath = torch.tensordot(op.to(bath.device), bath, ([2, 3], [1, 3]))
    bath = torch.tensordot(state.conj(), bath, ([1, 2], [1, 3]))
    return bath


"""
function to compute the right baths. The three indices in the bath are as follows:
(bond of state conj, bond of operator, bond of state)
The baths have shape
-xx
-xx
-xx
with the index ordering (top, middle, bottom)
bath tensors are put on the device of the factor to the left
"""


def right_baths(state: MPS, op: MPO, final_qubit: int) -> list[torch.Tensor]:
    state_factor = state.factors[-1]
    bath = torch.ones(1, 1, 1, device=state_factor.device, dtype=state_factor.dtype)
    baths = [bath]
    for i in range(len(state.factors) - 1, final_qubit - 1, -1):
        bath = new_right_bath(bath, state.factors[i], op.factors[i])
        bath = bath.to(state.factors[i - 1].device)
        baths.append(bath)
    return baths


"""
Computes H(psi) where
    x-    -x
    x  ||  x             ||
H = x- xx -x  and psi = -xx-
    x  ||  x
    x-    -x

Expects the two qubit factors of the MPS precontracted,
with one 'fat' physical index of dim 4 and index ordering
(left bond, physical index, right bond):
         ||
      -xxxxxx-
The Hamiltonian should have an index ordering of
(left bond, out, in, right bond).
The baths must have shape (top, middle, bottom).
All tensors must be on the same device
"""


def apply_effective_Hamiltonian(
    state: torch.Tensor,
    ham: torch.Tensor,
    left_bath: torch.Tensor,
    right_bath: torch.Tensor,
) -> torch.Tensor:
    assert left_bath.ndim == 3 and left_bath.shape[0] == left_bath.shape[2]
    assert right_bath.ndim == 3 and right_bath.shape[0] == right_bath.shape[2]
    assert left_bath.shape[2] == state.shape[0] and right_bath.shape[2] == state.shape[2]
    assert left_bath.shape[1] == ham.shape[0] and right_bath.shape[1] == ham.shape[3]

    # the optimal contraction order depends on the details
    # this order seems to be pretty balanced, but needs to be
    # revisited when use-cases are more well-known
    state = torch.tensordot(left_bath, state, 1)
    state = state.permute(0, 3, 1, 2)
    ham = ham.permute(0, 2, 1, 3)
    state = state.reshape(state.shape[0], state.shape[1], -1).contiguous()
    ham = ham.reshape(-1, ham.shape[2], ham.shape[3]).contiguous()
    state = torch.tensordot(state, ham, 1)
    state = state.permute(0, 2, 1, 3)
    state = state.reshape(state.shape[0], state.shape[1], -1).contiguous()
    right_bath = right_bath.permute(2, 1, 0)
    right_bath = right_bath.reshape(-1, right_bath.shape[2]).contiguous()
    state = torch.tensordot(state, right_bath, 1)
    return state


_TIME_CONVERSION_COEFF = 0.001  # Omega and delta are given in rad/μs, dt in ns


def evolve_pair(
    *,
    state_factors: list[torch.Tensor],
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factors: list[torch.Tensor],
    dt: float,
    orth_center_right: bool,
    is_hermitian: bool,
    config: MPSConfig,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Time evolution of a pair of tensors of a tensor train using baths and truncated SVD.
    Returned state tensors are kept on their respective devices.
    """
    assert len(state_factors) == 2
    assert len(baths) == 2
    assert len(ham_factors) == 2

    left_state_factor, right_state_factor = state_factors
    left_bath, right_bath = baths
    left_ham_factor, right_ham_factor = ham_factors

    left_device = left_state_factor.device
    right_device = right_state_factor.device

    left_bond_dim = left_state_factor.shape[0]
    right_bond_dim = right_state_factor.shape[-1]

    # Computation is done on left_device (arbitrary)

    combined_state_factors = torch.tensordot(
        left_state_factor, right_state_factor.to(left_device), dims=1
    ).reshape(left_bond_dim, 4, right_bond_dim)

    left_ham_factor = left_ham_factor.to(left_device)
    right_ham_factor = right_ham_factor.to(left_device)
    right_bath = right_bath.to(left_device)

    combined_hamiltonian_factors = (
        torch.tensordot(left_ham_factor, right_ham_factor, dims=1)
        .transpose(2, 3)
        .reshape(left_ham_factor.shape[0], 4, 4, -1)
    )

    op = (
        lambda x: -_TIME_CONVERSION_COEFF
        * 1j
        * dt
        * apply_effective_Hamiltonian(
            x, combined_hamiltonian_factors, left_bath, right_bath
        )
    )

    evol = krylov_exp(
        op,
        combined_state_factors,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    ).reshape(left_bond_dim * 2, 2 * right_bond_dim)

    l, r = split_tensor(
        evol,
        max_error=config.precision,
        max_rank=config.max_bond_dim,
        orth_center_right=orth_center_right,
    )

    return l.reshape(left_bond_dim, 2, -1), r.reshape(-1, 2, right_bond_dim).to(
        right_device
    )


def evolve_single(
    *,
    state_factor: torch.Tensor,
    baths: tuple[torch.Tensor, torch.Tensor],
    ham_factor: torch.Tensor,
    dt: float,
    is_hermitian: bool,
    config: MPSConfig,
) -> torch.Tensor:
    """
    Time evolution of a single tensor of a tensor train using baths.
    """
    assert len(baths) == 2

    left_bath, right_bath = baths

    op = (
        lambda x: -_TIME_CONVERSION_COEFF
        * 1j
        * dt
        * apply_effective_Hamiltonian(
            x,
            ham_factor,
            left_bath,
            right_bath,
        )
    )

    return krylov_exp(
        op,
        state_factor,
        exp_tolerance=config.precision * config.extra_krylov_tolerance,
        norm_tolerance=config.precision * config.extra_krylov_tolerance,
        max_krylov_dim=config.max_krylov_dim,
        is_hermitian=is_hermitian,
    )
